import importlib.metadata

import ase.io
import ase.visualize
import tqdm
import typer

app = typer.Typer()


@app.command()
def view(file: str):
    """Use ASE GUI to view a H5MD Trajectory."""
    import znh5md

    typer.echo(f"Loading atoms from {file}")

    io = znh5md.IO(file)
    ase.visualize.view(io[:])


@app.command()
def export(file: str, output: str):
    """Export a H5MD File into the output file."""
    import znh5md

    data = znh5md.IO(file)
    for atom in tqdm.tqdm(data[:], ncols=120):
        ase.io.write(output, atom, append=True)


@app.command()
def convert(file: str, db_file: str):
    """Save a trajectory as H5MD File."""
    import znh5md

    atoms = ase.io.read(file, ":")
    znh5md.write(db_file, atoms)


def version_callback(value: bool) -> None:
    """Get the installed dask4dvc version."""
    if value:
        typer.echo(f"znh5md {importlib.metadata.version('znh5md')}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    ),
) -> None:
    """ZnH5MD: A H5MD file viewer and converter."""
    _ = version  # this would be greyed out otherwise



import dataclasses
import enum
from typing import Any, Dict, List, Optional, TypedDict, Union

import h5py
import numpy as np
from ase import Atoms
from ase.calculators.calculator import all_properties

from .utils import concatenate_varying_shape_arrays


class ASEKeyMetaData(TypedDict):
    unit: Optional[str]
    calc: Optional[bool]


class CustomINFOData(str, enum.Enum):
    """Custom INFO data that is not stored in the particles group."""

    h5md_step = "step"
    h5md_time = "time"


UNIT_MAPPING = {
    "energy": "eV",
    "force": "eV/Angstrom",
    "stress": "eV/Angstrom^3",
    "velocity": "Angstrom/fs",
    "position": "Angstrom",
    "time": "fs",
}


@dataclasses.dataclass
class ASEData:
    """A dataclass for storing ASE Atoms data."""

    cell: Optional[np.ndarray]
    pbc: np.ndarray
    observables: Dict[str, np.ndarray]
    particles: Dict[str, np.ndarray]
    metadata: Optional[Dict[str, ASEKeyMetaData]] = None
    time: Optional[np.ndarray] = None
    step: Optional[np.ndarray] = None

    def __post_init__(self):
        if self.metadata is None:
            self.metadata = {}

        for key in UNIT_MAPPING:
            if key not in self.metadata:
                self.metadata[key] = {"unit": UNIT_MAPPING[key], "calc": False}

    def __len__(self):
        return len(self.particles["species"])


def get_property(group, name: str, prop: str, index) -> Optional[np.ndarray]:
    """Retrieve a property from an HDF5 group."""
    try:
        return group[name][prop]["value"][index]
    except KeyError:
        return None
    except OSError:
        # h5py raises an OSError if index == len(field)
        # but we need an IndexError
        raise IndexError(f"Index {index} out of bounds for property {prop}")


def get_atomic_numbers(group, name: str, index) -> Optional[np.ndarray]:
    """Retrieve atomic numbers from an HDF5 group."""
    return get_property(group, name, "species", index)


def get_positions(group, name: str, index) -> Optional[np.ndarray]:
    """Retrieve positions from an HDF5 group."""
    return get_property(group, name, "position", index)


def get_box(group, name: str, index) -> Optional[np.ndarray]:
    """Retrieve cell box dimensions from an HDF5 group."""
    return get_property(group, name, "box/edges", index)


def get_velocities(group, name: str, index) -> Optional[np.ndarray]:
    """Retrieve velocities from an HDF5 group."""
    return get_property(group, name, "velocity", index)


def get_pbc(group, name: str, index) -> np.ndarray:
    """Retrieve periodic boundary conditions from an HDF5 group."""
    if "pbc" in group[name]["box"]:
        return group[name]["box"]["pbc"]["value"][index]
    else:
        try:
            boundary = group[name]["box"].attrs["boundary"]
            return np.array([True if x == b"periodic" else False for x in boundary])
        except KeyError:
            return np.array([False, False, False])


def get_species_aux_data(
    group: h5py.Group, name: str, field: str, index: Union[int, list[int], slice]
) -> np.ndarray:
    """Helper function to retrieve data from an HDF5 group."""
    try:
        # Attempt to retrieve the data using the provided index
        data = group[name]["species"][field][index]
    except ValueError:
        # Backwards compatibility: Handle case where the data is stored as a scalar
        scalar_data = group[name]["species"][field][()]

        # Calculate the length of the indexed data without loading the actual data
        value_shape = group[name]["species"]["value"].shape
        if isinstance(index, slice):
            value_length = len(range(*index.indices(value_shape[0])))
            value_start = index.start if index.start is not None else 0
        elif isinstance(index, list):
            if not sorted(index) == index:
                raise ValueError("Indices must be sorted")
            value_length = len(index)
            value_start = index[0]
        elif isinstance(index, int):
            value_length = 1
            value_start = index
        else:
            raise TypeError("Unsupported index type")

        # data should be (start * scalar_data, end * scalar_data, step * scalar_data)
        data = np.arange(value_start, value_start + value_length) * scalar_data

    return data


def get_time(
    group: h5py.Group, name: str, index: Union[int, list[int], slice]
) -> np.ndarray:
    """Retrieve time from an HDF5 group."""
    return get_species_aux_data(group, name, "time", index)


def get_step(
    group: h5py.Group, name: str, index: Union[int, list[int], slice]
) -> np.ndarray:
    """Retrieve step from an HDF5 group."""
    return get_species_aux_data(group, name, "step", index)


ASE_TO_H5MD = {
    "numbers": "species",  # remove
    "positions": "position",  # remove
    "cell": "box",
}


def extract_atoms_data(atoms: Atoms, use_ase_calc: bool = True) -> ASEData:  # noqa: C901
    """
    Extract data from an ASE Atoms object and return an ASEData object.

    Args:
        atoms: ase.Atoms
            An ASE Atoms object containing the atomic structure.
        use_ase_calc: bool, optional
            Whether to include data from the ASE calculator.
            Defaults to True.

    Returns:
        ASEData:
            An object containing the extracted data, including particles'
            information, observables, metadata, time, and step.
    """
    atomic_numbers = atoms.get_atomic_numbers()
    positions = atoms.get_positions()
    cell = atoms.get_cell()
    pbc = atoms.get_pbc()
    velocities = atoms.get_velocities() if "momenta" in atoms.arrays else None

    particles: Dict[str, Any] = {
        "species": atomic_numbers,
        "position": positions,
    }
    if velocities is not None:
        particles["velocity"] = velocities

    info_data: Dict[str, Any] = {}
    uses_calc: List[str] = []

    if use_ase_calc and atoms.calc is not None:
        for key, value in atoms.calc.results.items():
            key = "force" if key == "forces" else key
            uses_calc.append(key)
            value = np.array(value) if isinstance(value, (int, float, list)) else value

            if value.ndim > 1 and value.shape[0] == len(atomic_numbers):
                particles[key] = value
            else:
                info_data[key] = value

    for key, value in atoms.info.items():
        if use_ase_calc and key in all_properties:
            raise ValueError(f"Key {key} is reserved for ASE calculator results.")
        if key not in ASE_TO_H5MD and key not in CustomINFOData.__members__:
            info_data[key] = value

    for key, value in atoms.arrays.items():
        if use_ase_calc and key in all_properties:
            raise ValueError(f"Key {key} is reserved for ASE calculator results.")
        if key not in ASE_TO_H5MD:
            particles[key] = value

    time: Optional[float] = atoms.info.get(CustomINFOData.h5md_time.name, None)
    step: Optional[int] = atoms.info.get(CustomINFOData.h5md_step.name, None)

    return ASEData(
        cell=cell,
        pbc=pbc,
        observables=info_data,
        particles=particles,
        metadata={
            key: {"unit": UNIT_MAPPING.get(key), "calc": True} for key in uses_calc
        },
        time=time,
        step=step,
    )


# TODO highlight that an additional dimension is added to ASEData here
def combine_asedata(data: List[ASEData]) -> ASEData:
    """Combine multiple ASEData objects into one."""
    cell = _combine_property([x.cell for x in data])
    pbc = np.array(
        [x.pbc if x.pbc is not None else [False, False, False] for x in data]
    )

    observables = _combine_dicts([x.observables for x in data])
    particles = _combine_dicts([x.particles for x in data])

    time_occurrences = sum([x.time is not None for x in data])
    step_occurrences = sum([x.step is not None for x in data])
    if time_occurrences == len(data):
        time = np.array([x.time for x in data])
    elif time_occurrences == 0:
        time = None
    else:
        raise ValueError("Time is not consistent across data objects")

    if step_occurrences == len(data):
        step = np.array([x.step for x in data])
    elif step_occurrences == 0:
        step = None
    else:
        raise ValueError("Step is not consistent across data objects")

    return ASEData(
        cell=cell,
        pbc=pbc,
        observables=observables,
        particles=particles,
        metadata=data[0].metadata,  # we assume they are all equal
        time=time,
        step=step,
    )


def _combine_property(properties: List[Optional[np.ndarray]]) -> Optional[np.ndarray]:
    """Helper function to combine varying shape arrays."""
    if all(x is None for x in properties):
        return None
    return concatenate_varying_shape_arrays(
        [x if x is not None else np.array([]) for x in properties]
    )


def _combine_dicts(dicts: List[Dict[str, np.ndarray]]) -> Dict[str, np.ndarray]:
    """Helper function to combine dictionaries containing numpy arrays."""
    combined = {}
    for key in dicts[0]:
        combined[key] = concatenate_varying_shape_arrays(
            [
                d[key] if isinstance(d[key], np.ndarray) else np.array(d[key])
                for d in dicts
            ]
        )
    return combined



import ase
import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator


def concatenate_varying_shape_arrays(arrays: list[np.ndarray]) -> np.ndarray:
    """Concatenate arrays of varying lengths into a numpy array.

    Gaps are filled with NaN values.

    Parameters
    ----------
        arrays (list of numpy.array): List of numpy arrays of varying lengths.

    Returns
    -------
        numpy.array: A numpy array where each row corresponds to an input array.

    Example:
    >>> concatenate_varying_shape_arrays([np.array([1, 2]), np.array([3, 4, 5])])
    array([[ 1.,  2., nan],
           [ 3.,  4.,  5.]])

    """
    if np.shape(arrays[0]) == ():
        return np.array(arrays).flatten()
    if len(np.shape(arrays[0])) == 0:
        return np.array(arrays)
    max_n_particles = max(x.shape[0] for x in arrays)
    dimensions = arrays[0].shape[1:]

    result = np.full((len(arrays), max_n_particles, *dimensions), np.nan)
    for i, x in enumerate(arrays):
        result[i, : x.shape[0], ...] = x
    return result


def remove_nan_rows(array: np.ndarray) -> np.ndarray | None:
    """Remove rows with NaN values from a numpy array.

    Parameters
    ----------
        array (numpy.array): At least 2D numpy array.

    Returns
    -------
        numpy.array: A numpy array where rows with NaN values are removed.

    Example:
    >>> remove_nan_rows(np.array([[ 1.,  2., np.nan], [ 3.,  4.,  5.]]))
    array([[3., 4., 5.]])
    >>> remove_nan_rows(np.nan)
    None
    >>> remove_nan_rows(np.array(1))
    1

    """
    if len(np.shape(array)) == 0:
        return array if not np.isnan(array) else None
    return array[~np.isnan(array).all(axis=tuple(range(1, array.ndim)))]


def fill_dataset(dataset, new_data):
    # Axis 0 is the configuration axis
    # Axis 1 is the number of particles axis
    # all following axis are optional, e.g 2 can be (x, y, z) or 1 can already be energy

    old_shape = dataset.shape
    new_shape = new_data.shape

    if len(old_shape) == 1 and len(new_shape) == 1:
        dataset.resize((old_shape[0] + new_shape[0],))
        dataset[old_shape[0] :] = new_data
        return

    # Determine the new shape of the dataset
    max_shape = (
        old_shape[0] + new_shape[0],
        max(old_shape[1], new_shape[1]),
        *old_shape[2:],
    )
    # raise ValueError(f"{old_shape=}, {new_shape=}, {max_shape=}")

    # Resize the dataset to the new shape
    dataset.resize(max_shape)

    # Fill the new columns of the existing data with np.nan
    if old_shape[1] < max_shape[1]:
        dataset[:, old_shape[1] :] = np.nan

    # Fill the new data rows with np.nan if necessary
    if new_shape[1] < max_shape[1]:
        padded_new_data = np.full((new_shape[0], max_shape[1], *old_shape[2:]), np.nan)
        padded_new_data[:, : new_shape[1]] = new_data
    else:
        padded_new_data = new_data

    # Append the new data to the dataset
    dataset[old_shape[0] :] = padded_new_data


def build_atoms(args) -> ase.Atoms:
    (
        atomic_numbers,
        positions,
        velocities,
        cell,
        pbc,
        calc_data,
        info_data,
        arrays_data,
    ) = args
    atomic_numbers = remove_nan_rows(atomic_numbers)
    if positions is not None:
        positions = remove_nan_rows(positions)
    if velocities is not None:
        velocities = remove_nan_rows(velocities)
    if calc_data is not None:
        calc_data = {key: remove_nan_rows(value) for key, value in calc_data.items()}
    if arrays_data is not None:
        arrays_data = {
            key: remove_nan_rows(value) for key, value in arrays_data.items()
        }
    if info_data is not None:
        info_data = {key: remove_nan_rows(value) for key, value in info_data.items()}

    atoms = ase.Atoms(
        symbols=atomic_numbers,
        positions=positions,
        velocities=velocities,
        pbc=pbc,
        cell=cell,
    )
    atoms.arrays.update(arrays_data)
    atoms.info.update(info_data)

    if calc_data is not None:
        if len(calc_data) > 0:
            atoms.calc = SinglePointCalculator(atoms=atoms)
            atoms.calc.results = calc_data

    return atoms


def build_structures(
    cell,
    pbc,
    arrays_data,
    calc_data,
    info_data,
) -> list[ase.Atoms]:
    structures = []

    positions = arrays_data.pop("positions", None)
    velocities = arrays_data.pop("velocity", None)
    atomic_numbers = arrays_data.pop("species")
    if atomic_numbers is not None:
        # could use ThreadPoolExecutor here
        # but there is no performance gain
        for idx in range(len(atomic_numbers)):
            args = (
                atomic_numbers[idx],
                positions[idx] if positions is not None else None,
                velocities[idx] if velocities is not None else None,
                cell[idx] if cell is not None else None,
                pbc[idx] if isinstance(pbc[0], np.ndarray) else pbc,
                {key: value[idx] for key, value in calc_data.items()},
                {key: value[idx] for key, value in info_data.items()},
                {key: value[idx] for key, value in arrays_data.items()},
            )
            structures.append(build_atoms(args))
    return structures



import contextlib
import dataclasses
import os
import pathlib
import typing as t
import warnings
from collections.abc import MutableSequence
from typing import List, Optional, Union

import ase
import h5py
import numpy as np
from ase.calculators.calculator import all_properties
from tqdm import tqdm

import znh5md.format as fmt
from znh5md import utils

# TODO: use pint to convert the units in the h5md file to ase units


@contextlib.contextmanager
def _open_file(
    filename: str | os.PathLike | None, file_handle: h5py.File | None, **kwargs
) -> t.Generator[h5py.File, None, None]:
    if file_handle is not None:
        yield file_handle
    else:
        with h5py.File(filename, **kwargs) as f:
            yield f


@dataclasses.dataclass
class IO(MutableSequence):
    """A class for handling H5MD files for ASE Atoms objects."""

    filename: Optional[str | os.PathLike] = None
    file_handle: Optional[h5py.File] = None
    pbc_group: bool = True  # Specify PBC per step (Not H5MD conform)
    save_units: bool = True  # Export ASE units into the H5MD file
    author: str = "N/A"
    author_email: str = "N/A"
    creator: str = "N/A"
    creator_version: str = "N/A"
    particle_group: Optional[str] = None
    compression: Optional[str] = "gzip"
    compression_opts: Optional[int] = None
    timestep: float = 1.0
    store: t.Literal["time", "linear"] = "linear"
    tqdm_limit: int = 100
    chunk_size: Optional[int] = None
    use_ase_calc: bool = True

    def __post_init__(self):
        if self.filename is None and self.file_handle is None:
            raise ValueError("Either filename or file_handle must be provided")
        if self.filename is not None and self.file_handle is not None:
            raise ValueError("Only one of filename or file_handle can be provided")
        if self.filename is not None:
            self.filename = pathlib.Path(self.filename)
        self._set_particle_group()
        self._read_author_creator()

    def _set_particle_group(self):
        if self.particle_group is not None:
            pass
        elif self.filename is not None and self.filename.exists():
            with _open_file(self.filename, self.file_handle, mode="r") as f:
                self.particle_group = next(iter(f["particles"].keys()))
        elif (
            self.file_handle is not None
            and pathlib.Path(self.file_handle.filename).exists()
        ):
            with _open_file(self.filename, self.file_handle, mode="r") as f:
                self.particle_group = next(iter(f["particles"].keys()))
        else:
            self.particle_group = "atoms"

    def _read_author_creator(self):
        with contextlib.suppress(FileNotFoundError, KeyError):
            # FileNotFoundError if the filename does not exist
            # KeyError if the file has not yet been initialized as H5MD
            #   or the keys are not provided, which is officially
            #   not allowed in H5MD.
            with _open_file(self.filename, self.file_handle, mode="r") as f:
                self.author = f["h5md"]["author"].attrs["name"]
                self.author_email = f["h5md"]["author"].attrs["email"]
                self.creator = f["h5md"]["creator"].attrs["name"]
                self.creator_version = f["h5md"]["creator"].attrs["version"]

    def create_file(self):
        with _open_file(self.filename, self.file_handle, mode="w") as f:
            g_h5md = f.create_group("h5md")
            g_h5md.attrs["version"] = np.array([1, 1])
            g_author = g_h5md.create_group("author")
            g_author.attrs["name"] = self.author
            g_author.attrs["email"] = self.author_email
            g_creator = g_h5md.create_group("creator")
            g_creator.attrs["name"] = self.creator
            g_creator.attrs["version"] = self.creator_version
            f.create_group("particles")

    def __len__(self) -> int:
        with _open_file(self.filename, self.file_handle, mode="r") as f:
            return len(f["particles"][self.particle_group]["species"]["value"])

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[ase.Atoms, List[ase.Atoms]]:
        single_item = isinstance(index, int)
        index = [index] if single_item else index

        arrays_data = {}
        calc_data = {}
        info_data = {}

        with _open_file(self.filename, self.file_handle, mode="r") as f:
            arrays_data["species"] = fmt.get_atomic_numbers(
                f["particles"], self.particle_group, index
            )
            positions = fmt.get_positions(f["particles"], self.particle_group, index)
            if positions is not None:
                arrays_data["positions"] = positions
            cell = fmt.get_box(f["particles"], self.particle_group, index)
            pbc = fmt.get_pbc(f["particles"], self.particle_group, index)

            self._extract_additional_data(f, index, arrays_data, calc_data, info_data)
            self._update_info_data_with_time_and_step(info_data, f, index)

        structures = utils.build_structures(
            cell,
            pbc,
            arrays_data,
            calc_data,
            info_data,
        )

        return structures[0] if single_item else structures

    def _update_info_data_with_time_and_step(self, info_data, f, index):
        # we use the time value from `species` because it is guaranteed to exist
        time = fmt.get_time(f["particles"], self.particle_group, index)
        step = fmt.get_step(f["particles"], self.particle_group, index)
        if any(key in info_data for key in fmt.CustomINFOData.__members__):
            raise ValueError("key is already present in the info data.")
        info_data[fmt.CustomINFOData.h5md_time.name] = time.tolist()
        info_data[fmt.CustomINFOData.h5md_step.name] = step.tolist()

    def _extract_additional_data(self, f, index, arrays_data, calc_data, info_data):
        for key in f["particles"][self.particle_group].keys():
            if key not in list(fmt.ASE_TO_H5MD.values()):
                if self.use_ase_calc and (
                    f["particles"][self.particle_group][key]["value"].attrs.get(
                        "ASE_CALCULATOR_RESULT"
                    )
                    or key in all_properties
                    or key == "force"
                ):
                    calc_data[key if key != "force" else "forces"] = fmt.get_property(
                        f["particles"], self.particle_group, key, index
                    )
                else:
                    arrays_data[key if key != "force" else "forces"] = fmt.get_property(
                        f["particles"], self.particle_group, key, index
                    )
        if f"observables/{self.particle_group}" in f:
            for key in f[f"observables/{self.particle_group}"].keys():
                if self.use_ase_calc and (
                    f["observables"][self.particle_group][key]["value"].attrs.get(
                        "ASE_CALCULATOR_RESULT"
                    )
                    or key in all_properties
                ):
                    calc_data[key] = fmt.get_property(
                        f["observables"], self.particle_group, key, index
                    )
                else:
                    info_data[key] = fmt.get_property(
                        f["observables"], self.particle_group, key, index
                    )

    def extend(self, images: List[ase.Atoms]):
        if len(images) == 0:
            warnings.warn("No data provided")
            return
        if self.filename is not None and not self.filename.exists():
            self.create_file()
        if self.file_handle is not None:
            needs_creation = False
            with _open_file(self.filename, self.file_handle, mode="r") as f:
                needs_creation = "h5md" not in f
            if needs_creation:
                self.create_file()

        data = []
        _disable_tqdm = len(images) < self.tqdm_limit if self.tqdm_limit > 0 else True
        for atoms in tqdm(
            images, ncols=120, desc="Preparing data", disable=_disable_tqdm
        ):
            data.append(fmt.extract_atoms_data(atoms, use_ase_calc=self.use_ase_calc))
        combined_data = fmt.combine_asedata(data)

        with _open_file(self.filename, self.file_handle, mode="a") as f:
            if self.particle_group not in f["particles"]:
                self._create_particle_group(f, combined_data)
            else:
                self._extend_existing_data(f, combined_data)

    def _create_particle_group(self, f, data: fmt.ASEData):
        g_particle_grp = f["particles"].create_group(self.particle_group)
        self._create_group(
            g_particle_grp, "box/edges", data.cell, time=data.time, step=data.step
        )
        g_particle_grp["box"].attrs["dimension"] = 3
        g_particle_grp["box"].attrs["boundary"] = [
            "periodic" if y else "none" for y in data.pbc[0]
        ]

        if self.pbc_group and data.pbc is not None:
            self._create_group(
                g_particle_grp, "box/pbc", data.pbc, time=data.time, step=data.step
            )

        _disable_tqdm = len(data) < self.tqdm_limit if self.tqdm_limit > 0 else True
        for key, value in tqdm(
            data.particles.items(),
            ncols=120,
            desc="Creating groups",
            disable=_disable_tqdm,
        ):
            self._create_group(
                g_particle_grp,
                key,
                value,
                data.metadata.get(key, {}).get("unit"),
                calc=data.metadata.get(key, {}).get("calc")
                if self.use_ase_calc
                else None,
                time=data.time,
                step=data.step,
            )
        self._create_observables(
            f,
            data.observables,
            data.metadata,
            time=data.time,
            step=data.step,
        )

    def _create_group(
        self,
        parent_grp,
        name,
        data,
        unit: Optional[str] = None,
        calc: Optional[bool] = None,
        time: np.ndarray | None = None,
        step: np.ndarray | None = None,
    ):
        if data is not None:
            g_grp = parent_grp.create_group(name)
            ds_value = g_grp.create_dataset(
                "value",
                data=data,
                dtype=np.float64,
                chunks=True
                if self.chunk_size is None
                else tuple([self.chunk_size] + list(data.shape[1:])),
                maxshape=([None] * data.ndim),
                compression=self.compression,
                compression_opts=self.compression_opts,
            )
            if calc is not None:
                ds_value.attrs["ASE_CALCULATOR_RESULT"] = calc
            if unit and self.save_units:
                ds_value.attrs["unit"] = unit
            if time is None:
                time = np.arange(len(data)) * self.timestep
            elif self.store == "linear":
                warnings.warn("time is ignored in 'linear' storage mode")
            if step is None:
                step = np.arange(len(data))
            elif self.store == "linear":
                warnings.warn("step is ignored in 'linear' storage mode")
            self._add_time_and_step(g_grp, step, time)

    def _add_time_and_step(self, grp, step: np.ndarray, time: np.ndarray):
        if self.store == "time":
            ds_time = grp.create_dataset(
                "time",
                dtype=np.float64,
                data=time,
                compression=self.compression,
                compression_opts=self.compression_opts,
                maxshape=(None,),
                chunks=True if self.chunk_size is None else (self.chunk_size,),
            )
            ds_time.attrs["unit"] = "fs"
            ds_step = grp.create_dataset(
                "step",
                dtype=int,
                data=step,
                compression=self.compression,
                compression_opts=self.compression_opts,
                maxshape=(None,),
                chunks=True if self.chunk_size is None else (self.chunk_size,),
            )
        elif self.store == "linear":
            ds_time = grp.create_dataset(
                "time",
                dtype=np.float64,
                data=self.timestep,
            )
            ds_time.attrs["unit"] = "fs"
            ds_step = grp.create_dataset(
                "step",
                dtype=int,
                data=1,
            )
            ds_time[()] = self.timestep
            ds_step[()] = 1

    def _create_observables(
        self,
        f,
        info_data,
        metadata: dict,
        time: np.ndarray | None = None,
        step: np.ndarray | None = None,
    ):
        if info_data:
            g_observables = f.require_group("observables")
            g_info = g_observables.require_group(self.particle_group)
            for key, value in info_data.items():
                g_observable = g_info.create_group(key)
                ds_value = g_observable.create_dataset(
                    "value",
                    data=value,
                    dtype=np.float64,
                    chunks=True
                    if self.chunk_size is None
                    else tuple([self.chunk_size] + list(value.shape[1:])),
                    maxshape=([None] * value.ndim),
                    compression=self.compression,
                    compression_opts=self.compression_opts,
                )
                if self.use_ase_calc and metadata.get(key, {}).get("calc") is not None:
                    ds_value.attrs["ASE_CALCULATOR_RESULT"] = metadata[key]["calc"]
                if metadata.get(key, {}).get("unit") and self.save_units:
                    ds_value.attrs["unit"] = metadata[key]["unit"]
                if time is None:
                    time = np.arange(len(value)) * self.timestep
                elif self.store == "linear":
                    warnings.warn("time is ignored in 'linear' storage mode")
                if step is None:
                    step = np.arange(len(value))
                elif self.store == "linear":
                    warnings.warn("step is ignored in 'linear' storage mode")
                self._add_time_and_step(g_observable, step, time)

    def _extend_existing_data(self, f, data: fmt.ASEData):
        g_particle_grp = f["particles"][self.particle_group]
        self._extend_group(
            g_particle_grp, "box/edges", data.cell, step=data.step, time=data.time
        )
        if self.pbc_group and data.pbc is not None:
            self._extend_group(
                g_particle_grp, "box/pbc", data.pbc, step=data.step, time=data.time
            )
        _disable_tqdm = len(data) < self.tqdm_limit if self.tqdm_limit > 0 else True
        for key, value in tqdm(
            data.particles.items(),
            ncols=120,
            desc="Extending groups",
            disable=_disable_tqdm,
        ):
            self._extend_group(
                g_particle_grp, key, value, step=data.step, time=data.time
            )
        self._extend_observables(f, data.observables, step=data.step, time=data.time)

    def _extend_group(
        self,
        parent_grp,
        name,
        data,
        step: np.ndarray | None = None,
        time: np.ndarray | None = None,
    ):
        if data is not None and name in parent_grp:
            g_grp = parent_grp[name]
            utils.fill_dataset(g_grp["value"], data)
            if self.store == "time":
                if time is None:
                    last_time = g_grp["time"][-1]
                    time = np.arange(len(data)) * self.timestep + last_time
                if step is None:
                    last_step = g_grp["step"][-1]
                    step = np.arange(len(data)) + last_step
                utils.fill_dataset(
                    g_grp["time"],
                    time,
                )
                utils.fill_dataset(g_grp["step"], step)

    def _extend_observables(
        self,
        f,
        info_data,
        step: np.ndarray | None = None,
        time: np.ndarray | None = None,
    ):
        if f"observables/{self.particle_group}" in f:
            g_observables = f[f"observables/{self.particle_group}"]
            for key, value in info_data.items():
                if key in g_observables:
                    g_val = g_observables[key]
                    utils.fill_dataset(g_val["value"], value)
                    if self.store == "time":
                        if time is None:
                            last_time = g_val["time"][-1]
                            time = np.arange(len(value)) * self.timestep + last_time
                        if step is None:
                            last_step = g_val["step"][-1]
                            step = np.arange(len(value)) + last_step
                        utils.fill_dataset(
                            g_val["time"],
                            time,
                        )
                        utils.fill_dataset(g_val["step"], step)

    def append(self, atoms: ase.Atoms):
        self.extend([atoms])

    def __delitem__(self, index):
        raise NotImplementedError("Deleting items is not supported")

    def __setitem__(self, index, value):
        raise NotImplementedError("Setting items is not supported")

    def insert(self, index, value):
        raise NotImplementedError("Inserting items is not supported")
